package org.bitbucket.bradleysmithllc.etlunit.io.file;

import org.apache.commons.io.IOUtils;

import java.io.Reader;
import java.io.StringWriter;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.sql.Clob;
import java.sql.Types;
import java.text.SimpleDateFormat;

public class JdbcTypeConverter implements DataFile.DataConverter
{
	// use this formatter to ensure that all databases represent nanos uniformly
	// (they are truncated)
	private final SimpleDateFormat timestampFormatter = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss.SSS");

	@Override
	public String format(Object data, DataFileSchema.Column column)
	{
		String columnType = column.getType();
		int type = Types.VARCHAR;

		// if the type is null, varchar is the default
		if (columnType != null)
		{
			type = getTypeValue(columnType).intValue();
		}

		// grab the corresponding type code for this id
		switch (type)
		{
			case Types.TIMESTAMP:
				return timestampFormatter.format(data);
			case Types.INTEGER:
			case Types.SMALLINT:
			case Types.BIGINT:
			case Types.TINYINT:
			case Types.CHAR:
			case Types.LONGVARCHAR:
			case Types.VARCHAR:
			case Types.DATE:
			case Types.TIME:

			case Types.BIT:
			case Types.FLOAT:
			case Types.DECIMAL:
			case Types.DOUBLE:
			case Types.NUMERIC:
			case Types.REAL:
			case Types.BOOLEAN:
				return String.valueOf(data);
			case Types.CLOB:
				if (data instanceof Clob)
				{
					Clob clob = (Clob) data;

					try
					{
						Reader reader = clob.getCharacterStream();

						try
						{
							StringWriter stw = new StringWriter();
							IOUtils.copy(reader, stw);

							return stw.toString();
						}
						finally
						{
							reader.close();
						}
					}
					catch (Exception e)
					{
						throw new RuntimeException("Error processing CLOB", e);
					}
				}
				else
				{
					return String.valueOf(data);
				}
			default:
				throw new UnsupportedOperationException("SQL data type " + columnType + " not supported");
		}
	}

	public static Integer getTypeValue(String name)
	{
		try
		{
			Field field = Types.class.getDeclaredField(name);

			Object value = field.get(null);

			if (value instanceof Integer)
			{
				return (Integer) value;
			}
			else
			{
				throw new IllegalArgumentException("Bad column type name: "
						+ name
						+ ", not found in Types.class as an integer field");
			}
		}
		catch (NoSuchFieldException e)
		{
			throw new IllegalArgumentException("Colummn type '" + name + "' not found");
		}
		catch (IllegalAccessException e)
		{
			throw new IllegalArgumentException("Colummn type '" + name + "' not accessible");
		}
	}

	public static String getTypeName(Integer code)
	{
		try
		{
			Field[] fields = Types.class.getDeclaredFields();

			for (int i = 0; i < fields.length; i++)
			{
				int modifiers = fields[i].getModifiers();
				if (Modifier.isStatic(modifiers) && Modifier.isPublic(modifiers))
				{
					// grab the value and see if it is an Integer.  If so, compare
					Object val = fields[i].get(null);

					if (val instanceof Integer)
					{
						if (((Integer) val).intValue() == code)
						{
							// this is the one
							return fields[i].getName();
						}
					}
				}
			}
		}
		catch (IllegalAccessException e1)
		{
			throw new IllegalArgumentException("", e1);
		}

		throw new IllegalArgumentException("Sql type code " + code + " not found in java.sql.Types.class");
	}
}